@@ -124,23 +124,25 @@ made initially (you can always re-do this but it would take some time) include:
 do you only want LRGs or do you want other tracers as well? 
 Do you want to enable satellite profile flexibilities (the :math:`s, s_v, s_p, s_r`
 parameters)? If so, you need to turn on ``want_ranks`` flag in the config file. 
-If you want to enable secondary bias based on local environment, what scale 
-radius do you want the environment do be defined in, this is set by the 
-``density_sigma`` flag in Mpc/h. The default value is 3. Related, the ``Ndim``
-parameter sets the grid size used to compute local density, and it should be set
-to be larger than Lbox/sigma_density. 
+If you want to enable secondary bias, you need to set ``want_AB`` flag to true in the
+config file. The local environment is defined by total mass within 5 Mpc/h but beyond
+``r98``. 
 
-Now you need to run the ``prepare_sim`` script, this extracts the simulation outputs
+IMPORTANT: Running this code is a two-part process. First, you need to run the ``prepare_sim``
+code, which generates the necessary data files for that simulation. Then you can run the actual
+HOD code. The first step only needs to be done once for a simulation box, but it can be slow, 
+depending on the downsampling and the features you choose to enable. 
+
+So first, you need to run the ``prepare_sim`` script, this extracts the simulation outputs
 and organizes them into formats that are suited for the HOD code. This code can take 
 approximately an hour depending on your configuration settings and system capabilities. 
-We recommend setting the ``Nthread_load`` parameter to ``min(sys_core_count, memoryGB_divided_by_20)``.
+We recommend setting the ``Nthread_load`` parameter to ``min(sys_core_count, memoryGB_divided_by_30)``.
 You can run ``load_sims`` on command line with ::
     python -m abacusnbody.hod.prepare_sim --path2config PATH2CONFIG
 
-Within Python, you can run the same script with
 
->>> from abacusnbody.hod import prepare_sim
->>> prepare_sim.main(/path/to/config.yaml)
+Within Python, you can run the same script with ``from abacusnbody.hod import prepare_sim``
+and then ``prepare_sim.main(/path/to/config.yaml)``.
 
 If your config file lives in the default location, i.e. ``./config``, then you 
 can ignore the ``-path2config`` flag. 
@@ -154,55 +156,53 @@ simply run the given script in bash ::
     python run_hod.py --path2config PATH2CONFIG
 
 You can also consruct the AbacusHOD object yourself within Python and run HODs from
-there. Here we show the scripts within ``run_hod.py`` for reference.
+there. Here we show the scripts within ``run_hod.py`` for reference.::
+    import os
+    import glob
+    import time
+    import yaml
+    import numpy as np
+    import argparse
 
->>> import os
->>> import glob
->>> import time
->>> 
->>> import yaml
->>> import numpy as np
->>> import argparse
->>> 
->>> from abacusnbody.hod.abacus_hod import AbacusHOD
->>> 
->>> path2config = 'config/abacus_hod.yaml' # path to config file
->>> 
->>> # load the config file and parse in relevant parameters
->>> config = yaml.load(open(path2config))
->>> sim_params = config['sim_params']
->>> HOD_params = config['HOD_params']
->>> clustering_params = config['clustering_params']
->>> 
->>> # additional parameter choices
->>> want_rsd = HOD_params['want_rsd']
->>> write_to_disk = HOD_params['write_to_disk']
->>> 
->>> # create a new AbacusHOD object
->>> newBall = AbacusHOD(sim_params, HOD_params, clustering_params)
->>>     
->>> # first hod run, slow due to compiling jit, write to disk
->>> mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk, Nthread = 16)
->>> 
->>> # run the 10 different HODs for timing
->>> for i in range(10):
->>>     newBall.tracers['LRG']['alpha'] += 0.01
->>>     print("alpha = ",newBall.tracers['LRG']['alpha'])
->>>     start = time.time()
->>>     mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk = False, Nthread = 64)
->>>     print("Done iteration ", i, "took time ", time.time() - start)
+    from abacusnbody.hod.abacus_hod import AbacusHOD
+
+    path2config = 'config/abacus_hod.yaml' # path to config file
+
+    # load the config file and parse in relevant parameters
+    config = yaml.safe_load(open(path2config))
+    sim_params = config['sim_params']
+    HOD_params = config['HOD_params']
+    clustering_params = config['clustering_params']
+
+    # additional parameter choices
+    want_rsd = HOD_params['want_rsd']
+    write_to_disk = HOD_params['write_to_disk']
+
+    # create a new AbacusHOD object
+    newBall = AbacusHOD(sim_params, HOD_params, clustering_params)
+        
+    # first hod run, slow due to compiling jit, write to disk
+    mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk, Nthread = 16)
+
+    # run the 10 different HODs for timing
+    for i in range(10):
+        newBall.tracers['LRG']['alpha'] += 0.01
+        print("alpha = ",newBall.tracers['LRG']['alpha'])
+        start = time.time()
+        mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk = False, Nthread = 64)
+        print("Done iteration ", i, "took time ", time.time() - start)
 
 The class also provides fast 2PCF calculators. For example to compute the 
-redshift-space 2PCF (:math:`\\xi(r_p, \\pi)`):
+redshift-space 2PCF (:math:`\\xi(r_p, \\pi)`): ::
 
->>> # load the rp pi binning from the config file
->>> bin_params = clustering_params['bin_params']
->>> rpbins = np.logspace(bin_params['logmin'], bin_params['logmax'], bin_params['nbins'])
->>> pimax = clustering_params['pimax']
->>> pi_bin_size = clustering_params['pi_bin_size']    # the pi binning is configrured by pi_max and bin size
->>> 
->>> mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk)
->>> xirppi = newBall.compute_xirppi(mock_dict, rpbins, pimax, pi_bin_size)
+    # load the rp pi binning from the config file
+    bin_params = clustering_params['bin_params']
+    rpbins = np.logspace(bin_params['logmin'], bin_params['logmax'], bin_params['nbins'])
+    pimax = clustering_params['pimax']
+    pi_bin_size = clustering_params['pi_bin_size']    # the pi binning is configrured by pi_max and bin size
+
+    mock_dict = newBall.run_hod(newBall.tracers, want_rsd, write_to_disk)
+    xirppi = newBall.compute_xirppi(mock_dict, rpbins, pimax, pi_bin_size)
 
 """
 import os
@@ -219,9 +219,8 @@ import multiprocessing
 from multiprocessing import Pool
 from astropy.io import ascii
 
-
-from .GRAND_HOD import gen_gal_cat
-from .tpcf_corrfunc import calc_xirppi_fast, calc_wp_fast
+from .GRAND_HOD import *
+from .tpcf_corrfunc import calc_xirppi_fast, calc_wp_fast, calc_multipole_fast
 # TODO B.H.: staging can be shorter and prettier; perhaps asdf for h5 and ecsv?
 
 class AbacusHOD:
@@ -243,10 +242,9 @@ class AbacusHOD:
             Dictionary of simulation parameters. Load from ``config/abacus_hod.yaml``. The dictionary should contain the following keys:
                 * ``sim_name``: str, name of the simulation volume, e.g. 'AbacusSummit_base_c000_ph006'. 
                 * ``sim_dir``: str, the directory that the simulation lives in, e.g. '/path/to/AbacusSummit/'.                                 
-                * ``scratch_dir``: str, the diretory to save galaxy to, e.g. '/my/output/galalxies'. 
+                * ``output_dir``: str, the diretory to save galaxy to, e.g. '/my/output/galalxies'. 
                 * ``subsample_dir``: str, where to save halo+particle subsample, e.g. '/my/output/subsamples/'. 
                 * ``z_mock``: float, which redshift slice, e.g. 0.5.    
-                * ``Nthread_load``: int, how many threads to use to load the simulation data, default 7. 
 
         HOD_params: dict 
             HOD parameters and tracer configurations. Load from ``config/abacus_hod.yaml``. It contains the following keys:
@@ -279,7 +277,7 @@ class AbacusHOD:
         self.sim_dir = sim_params['sim_dir']
         self.subsample_dir = sim_params['subsample_dir']
         self.z_mock = sim_params['z_mock']
-        self.scratch_dir = sim_params['scratch_dir']
+        self.output_dir = sim_params['output_dir']
         
         # tracers
         tracer_flags = HOD_params['tracer_flags']
@@ -292,7 +290,7 @@ class AbacusHOD:
         # HOD parameter choices
         self.want_ranks = HOD_params['want_ranks']
         self.want_rsd = HOD_params['want_rsd']
-        
+
         # clusteringparameters
         self.pimax = clustering_params['pimax']
         self.pi_bin_size = clustering_params['pi_bin_size']
@@ -303,16 +301,28 @@ class AbacusHOD:
         # load the subsample particles
         self.halo_data, self.particle_data, self.params, self.mock_dir = self.staging()
 
+        # determine the halo mass function
+        self.logMbins = np.linspace(
+            np.log10(np.min(self.halo_data['hmass'])), 
+            np.log10(np.max(self.halo_data['hmass'])), 101)
+        self.deltacbins = np.linspace(-0.5, 0.5, 101)
+        self.fenvbins = np.linspace(-0.5, 0.5, 101)
+
+        self.halo_mass_func, edges = np.histogramdd(
+            np.vstack((np.log10(self.halo_data['hmass']), self.halo_data['hdeltac'], self.halo_data['hfenv'])).T,
+            bins = [self.logMbins, self.deltacbins, self.fenvbins],
+            weights = self.halo_data['hmultis'])
+
 
     def staging(self):
         """
         Constructor call this function to load the halo+particle subsamples onto memory. 
         """
         # all paths relevant for mock generation
-        scratch_dir = Path(self.scratch_dir)
+        output_dir = Path(self.output_dir)
         simname = Path(self.sim_name)
         sim_dir = Path(self.sim_dir)
-        mock_dir = scratch_dir / simname / ('z%4.3f'%self.z_mock)
+        mock_dir = output_dir / simname / ('z%4.3f'%self.z_mock)
         # create mock_dir if not created
         mock_dir.mkdir(parents = True, exist_ok = True)
         subsample_dir = \
@@ -333,19 +343,6 @@ class AbacusHOD:
         params['velz2kms'] = header['VelZSpace_to_kms']/params['Lbox']
         params['numslabs'] = len(halo_info_fns)
         self.lbox = header['BoxSize']
-        
-        # # figure out number of halos and particles
-        # if 'ELG' not in self.tracers.keys() and 'QSO' not in self.tracers.keys():
-        #     numfile_name = subsample_dir / ("num_halos_parts.txt")
-        # else:
-        #     numfile_name = subsample_dir / ("num_halos_parts_MT.txt")
-        # fnum = np.loadtxt(numfile_name)
-        # fnum = np.array(sorted(fnum, key=lambda x : x[0])).astype(int)
-        # ichunks = fnum[:, 0]
-        # Nhalos = fnum[:, 1]
-        # Nparts = fnum[:, 2]
-        # Nhalos_tot = np.sum(Nhalos)
-        # Nparts_tot = np.sum(Nparts)
 
         # count ther number of halos and particles
         Nhalos = np.empty(params['numslabs'])
@@ -353,8 +350,8 @@ class AbacusHOD:
         for eslab in range(params['numslabs']):
             
             if 'ELG' not in self.tracers.keys() and 'QSO' not in self.tracers.keys():
-                halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod'%eslab)
-                particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod'%eslab)
+                halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod_oldfenv'%eslab)
+                particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod_oldfenv'%eslab)
             else:
                 halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod_MT'%eslab)
                 particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod_MT'%eslab)            
@@ -410,8 +407,8 @@ class AbacusHOD:
             print("Loading simulation by slab, ", eslab)
             
             if 'ELG' not in self.tracers.keys() and 'QSO' not in self.tracers.keys():
-                halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod'%eslab)
-                particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod'%eslab)
+                halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod_oldfenv'%eslab)
+                particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod_oldfenv'%eslab)
             else:
                 halofilename = subsample_dir / ('halos_xcom_%d_seed600_abacushod_MT'%eslab)
                 particlefilename = subsample_dir / ('particles_xcom_%d_seed600_abacushod_MT'%eslab)            
@@ -421,7 +418,6 @@ class AbacusHOD:
             halofilename = str(halofilename) + '_new.h5'
             particlefilename = str(particlefilename) + '_new.h5'
 
-            start = time.time()
             newfile = h5py.File(halofilename, 'r')
             maskedhalos = newfile['halos']
 
@@ -438,7 +434,6 @@ class AbacusHOD:
             halo_multi = maskedhalos['multi_halos']
             halo_submask = maskedhalos['mask_subsample'].astype(bool)
             halo_randoms = maskedhalos['randoms']
-            print("loading halo slab took ", time.time() - start)
 
             hpos[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_pos
             hvel[halo_ticker: halo_ticker + Nhalos[eslab]] = halo_vels
@@ -518,11 +513,11 @@ class AbacusHOD:
             particle_data['pranksv'] =  np.ones(Nparts_tot)
             particle_data['pranksp'] =  np.ones(Nparts_tot)
             particle_data['pranksr'] =  np.ones(Nparts_tot)
-            
+        
         return halo_data, particle_data, params, mock_dir
 
     
-    def run_hod(self, tracers, want_rsd = True, write_to_disk = False, Nthread = 16):
+    def run_hod(self, tracers = None, want_rsd = True, write_to_disk = False, Nthread = 16, verbose = False):
         """
         Runs a custom HOD.
 
@@ -540,7 +535,10 @@ class AbacusHOD:
             output to disk? default ``False``. Setting to ``True`` decreases performance. 
 
         ``Nthread``: int
-            Number of threads in the HOD run. Default 16. 
+            number of threads in the HOD run. Default 16. 
+
+        ``verbose``: bool, 
+            detailed stdout? default ``False``.
 
         Returns
         -------
@@ -555,11 +553,155 @@ class AbacusHOD:
             in the catalog are always centrals and the rest are satellites. 
 
         """
-        mock_dict = gen_gal_cat(self.halo_data, self.particle_data, self.tracers, self.params, Nthread,
-            enable_ranks = self.want_ranks, rsd = want_rsd, write_to_disk = write_to_disk, savedir = self.mock_dir)
+        if tracers == None:
+            tracers = self.tracers
+
+        mock_dict = gen_gal_cat(self.halo_data, self.particle_data, tracers, self.params, Nthread, 
+            enable_ranks = self.want_ranks, 
+            rsd = want_rsd, 
+            write_to_disk = write_to_disk, 
+            savedir = self.mock_dir,
+            verbose = False)
 
         return mock_dict
 
+    def compute_ngal(self, tracers = None, Nthread = 16):
+        """
+        Computes the number of each tracer generated by the HOD
+
+        Parameters
+        ----------
+        ``tracers``: dict
+            dictionary of multi-tracer HOD. ``tracers['LRG']`` is the dictionary of LRG HOD parameters,
+            overwrites the ``LRG_params`` argument in the constructor.
+            Same for keys ``'ELG'`` and ``'QSO'``.
+
+        ``Nthread``: int
+            Number of threads in the HOD run. Default 16. 
+
+        Returns
+        -------
+        ngal_dict: dict
+        dictionary of number of each tracer. 
+
+        fsat_dict: dict
+        dictionary of satellite fraction of each tracer.
+
+        """
+        if tracers == None:
+            tracers = self.tracers
+
+        ngal_dict = {}
+        fsat_dict = {}
+        for etracer in tracers.keys():
+            tracer_hod = tracers[etracer]
+            if etracer == 'LRG':
+                newngal = AbacusHOD._compute_ngal_lrg(
+                    self.logMbins, self.deltacbins, self.fenvbins, self.halo_mass_func,
+                    tracer_hod['logM_cut'], tracer_hod['logM1'], tracer_hod['sigma'], 
+                    tracer_hod['alpha'], tracer_hod['kappa'], tracer_hod['Acent'], 
+                    tracer_hod['Asat'], tracer_hod['Bcent'], tracer_hod['Bsat'], tracer_hod['ic'], Nthread)
+                ngal_dict[etracer] = newngal[0] + newngal[1]
+                fsat_dict[etracer] = newngal[1] / (newngal[0] + newngal[1])
+            elif etracer == 'ELG':
+                newngal = AbacusHOD._compute_ngal_elg(
+                    self.logMbins, self.deltacbins, self.fenvbins, self.halo_mass_func, 
+                    tracer_hod['p_max'], tracer_hod['Q'], tracer_hod['logM_cut'], 
+                    tracer_hod['kappa'], tracer_hod['sigma'], tracer_hod['logM1'],  
+                    tracer_hod['alpha'], tracer_hod['gamma'], tracer_hod['A_s'],  
+                    tracer_hod['Acent'],  tracer_hod['Asat'],  tracer_hod['Bcent'],  tracer_hod['Bsat'], Nthread) 
+                ngal_dict[etracer] = newngal[0] + newngal[1]
+                fsat_dict[etracer] = newngal[1] / (newngal[0] + newngal[1])
+            elif etracer == 'QSO':
+                newngal = AbacusHOD._compute_ngal_qso(
+                    self.logMbins, self.deltacbins, self.fenvbins, self.halo_mass_func, 
+                    tracer_hod['p_max'], tracer_hod['logM_cut'], 
+                    tracer_hod['kappa'], tracer_hod['sigma'], tracer_hod['logM1'],  
+                    tracer_hod['alpha'], tracer_hod['A_s'],  
+                    tracer_hod['Acent'],  tracer_hod['Asat'],  tracer_hod['Bcent'],  tracer_hod['Bsat'], Nthread)         
+                ngal_dict[etracer] = newngal[0] + newngal[1]
+                fsat_dict[etracer] = newngal[1] / (newngal[0] + newngal[1])
+        return ngal_dict, fsat_dict
+
+    @staticmethod
+    @njit(fastmath = True, parallel = True)
+    def _compute_ngal_lrg(logMbins, deltacbins, fenvbins, halo_mass_func,
+                   logM_cut, logM1, sigma, alpha, kappa, Acent, Asat, Bcent, Bsat, ic, Nthread):
+        """
+        internal helper to computer number of LRGs
+        """
+        numba.set_num_threads(Nthread)
+
+        logMs = 0.5*(logMbins[1:] + logMbins[:-1])
+        deltacs = 0.5*(deltacbins[1:] + deltacbins[:-1])
+        fenvs = 0.5*(fenvbins[1:] + fenvbins[:-1])
+        ngal_cent = 0
+        ngal_sat = 0
+        for i in numba.prange(len(logMbins) - 1):
+            for j in range(len(deltacbins) - 1):
+                for k in range(len(fenvbins) - 1):
+                    Mh_temp = 10**logMs[i]
+                    logM_cut_temp = logM_cut + Acent * deltacs[j] + Bcent * fenvs[k]
+                    M1_temp = 10**(logM1 + Asat * deltacs[j] + Bsat * fenvs[k])
+                    ncent_temp = n_cen_LRG(Mh_temp, logM_cut_temp, sigma)
+                    nsat_temp = n_sat_LRG_modified(Mh_temp, logM_cut_temp, 
+                        10**logM_cut_temp, M1_temp, sigma, alpha, kappa)
+                    ngal_cent += halo_mass_func[i, j, k] * ncent_temp * ic
+                    ngal_sat += halo_mass_func[i, j, k] * nsat_temp * ic
+        return ngal_cent, ngal_sat
+
+    @staticmethod
+    @njit(fastmath = True, parallel = True)
+    def _compute_ngal_elg(logMbins, deltacbins, fenvbins, halo_mass_func, p_max, Q,
+                   logM_cut, kappa, sigma, logM1, alpha, gamma, A_s, Acent, Asat, Bcent, Bsat, Nthread):
+        """
+        internal helper to computer number of LRGs
+        """
+        numba.set_num_threads(Nthread)
+
+        logMs = 0.5*(logMbins[1:] + logMbins[:-1])
+        deltacs = 0.5*(deltacbins[1:] + deltacbins[:-1])
+        fenvs = 0.5*(fenvbins[1:] + fenvbins[:-1])
+        ngal_cent = 0
+        ngal_sat = 0
+        for i in numba.prange(len(logMbins) - 1):
+            for j in range(len(deltacbins) - 1):
+                for k in range(len(fenvbins) - 1):
+                    Mh_temp = 10**logMs[i]
+                    logM_cut_temp = logM_cut + Acent * deltacs[j] + Bcent * fenvs[k]
+                    M1_temp = 10**(logM1 + Asat * deltacs[j] + Bsat * fenvs[k])
+                    ncent_temp = N_cen_ELG_v1(Mh_temp, p_max, Q, logM_cut_temp, sigma, gamma)
+                    nsat_temp = N_sat_generic(Mh_temp, 10**logM_cut_temp, kappa, M1_temp, alpha, A_s)
+                    ngal_cent += halo_mass_func[i, j, k] * ncent_temp
+                    ngal_sat += halo_mass_func[i, j, k] * nsat_temp
+        return ngal_cent, ngal_sat
+
+    @staticmethod
+    @njit(fastmath = True, parallel = True)
+    def _compute_ngal_qso(logMbins, deltacbins, fenvbins, halo_mass_func, p_max,
+                   logM_cut, kappa, sigma, logM1, alpha, A_s, Acent, Asat, Bcent, Bsat, Nthread):
+        """
+        internal helper to computer number of LRGs
+        """
+        numba.set_num_threads(Nthread)
+
+        logMs = 0.5*(logMbins[1:] + logMbins[:-1])
+        deltacs = 0.5*(deltacbins[1:] + deltacbins[:-1])
+        fenvs = 0.5*(fenvbins[1:] + fenvbins[:-1])
+        ngal_cent = 0
+        ngal_sat = 0
+        for i in numba.prange(len(logMbins) - 1):
+            for j in range(len(deltacbins) - 1):
+                for k in range(len(fenvbins) - 1):
+                    Mh_temp = 10**logMs[i]
+                    logM_cut_temp = logM_cut + Acent * deltacs[j] + Bcent * fenvs[k]
+                    M1_temp = 10**(logM1 + Asat * deltacs[j] + Bsat * fenvs[k])
+                    ncent_temp = N_cen_QSO(Mh_temp, p_max, logM_cut_temp, sigma)
+                    nsat_temp = N_sat_generic(Mh_temp, 10**logM_cut_temp, kappa, M1_temp, alpha, A_s)
+                    ngal_cent += halo_mass_func[i, j, k] * ncent_temp * ic
+                    ngal_sat += halo_mass_func[i, j, k] * nsat_temp * ic
+        return ngal_cent, ngal_sat
+
     def compute_clustering(self, mock_dict, *args, **kwargs):
         """
         Computes summary statistics, currently enabling ``wp`` and ``xirppi``.
@@ -592,6 +734,8 @@ class AbacusHOD:
             clustering = self.compute_xirppi(mock_dict, *args, **kwargs)
         elif self.clustering_type == 'wp':
             clustering = self.compute_wp(mock_dict, *args, **kwargs)
+        elif self.clustering_type == 'multipole':
+            clustering = self.compute_multipole(mock_dict, *args, **kwargs)
         return clustering
     
     def compute_xirppi(self, mock_dict, rpbins, pimax, pi_bin_size, Nthread = 8):
@@ -629,12 +773,41 @@ class AbacusHOD:
             z1 = mock_dict[tr1]['z']
             for i2, tr2 in enumerate(mock_dict.keys()):
                 if i1 > i2: continue # cross-correlations are symmetric
-                x2 = mock_dict[tr2]['x']
-                y2 = mock_dict[tr2]['y']
-                z2 = mock_dict[tr2]['z']
-                clustering[tr1+'_'+tr2] = calc_xirppi_fast(x1, y1, z1, rpbins, pimax, pi_bin_size, 
-                    self.lbox, Nthread, x2 = x2, y2 = y2, z2 = z2)
-                if i1 != i2: clustering[tr2+'_'+tr1] = clustering[tr1+'_'+tr2]
+                if i1 == i2: # auto corr
+                    clustering[tr1+'_'+tr2] = calc_xirppi_fast(x1, y1, z1, rpbins, pimax, pi_bin_size, 
+                        self.lbox, Nthread)
+                else:
+                    x2 = mock_dict[tr2]['x']
+                    y2 = mock_dict[tr2]['y']
+                    z2 = mock_dict[tr2]['z']
+                    clustering[tr1+'_'+tr2] = calc_xirppi_fast(x1, y1, z1, rpbins, pimax, pi_bin_size, 
+                        self.lbox, Nthread, x2 = x2, y2 = y2, z2 = z2)
+                    clustering[tr2+'_'+tr1] = clustering[tr1+'_'+tr2]
+        return clustering
+
+    def compute_multipole(self, mock_dict, rpbins, pimax, Nthread = 8):
+        clustering = {}
+        for i1, tr1 in enumerate(mock_dict.keys()):
+            x1 = mock_dict[tr1]['x']
+            y1 = mock_dict[tr1]['y']
+            z1 = mock_dict[tr1]['z']
+            for i2, tr2 in enumerate(mock_dict.keys()):
+                if i1 > i2: continue # cross-correlations are symmetric
+                if i1 == i2: # auto corr
+                    new_multi = calc_multipole_fast(x1, y1, z1, rpbins,
+                        self.lbox, Nthread)
+                    new_wp = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread)
+                    clustering[tr1+'_'+tr2] = np.concatenate((new_wp, new_multi))
+                else:
+                    x2 = mock_dict[tr2]['x']
+                    y2 = mock_dict[tr2]['y']
+                    z2 = mock_dict[tr2]['z']
+                    new_multi = calc_multipole_fast(x1, y1, z1, rpbins,
+                        self.lbox, Nthread, x2 = x2, y2 = y2, z2 = z2)
+                    new_wp = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread, 
+                        x2 = x2, y2 = y2, z2 = z2)
+                    clustering[tr1+'_'+tr2] = np.concatenate((new_wp, new_multi))
+                    clustering[tr2+'_'+tr1] = clustering[tr1+'_'+tr2]
         return clustering
 
     def compute_wp(self, mock_dict, rpbins, pimax, pi_bin_size, Nthread = 8):
@@ -672,14 +845,18 @@ class AbacusHOD:
             z1 = mock_dict[tr1]['z']
             for i2, tr2 in enumerate(mock_dict.keys()):
                 if i1 > i2: continue # cross-correlations are symmetric
-                x2 = mock_dict[tr2]['x']
-                y2 = mock_dict[tr2]['y']
-                z2 = mock_dict[tr2]['z']
-                clustering[tr1+'_'+tr2] = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread, x2 = x2, y2 = y2, z2 = z2)
-                if i1 != i2: clustering[tr2+'_'+tr1] = clustering[tr1+'_'+tr2]
+                if i1 == i2:
+                    clustering[tr1+'_'+tr2] = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread)
+                else:
+                    x2 = mock_dict[tr2]['x']
+                    y2 = mock_dict[tr2]['y']
+                    z2 = mock_dict[tr2]['z']
+                    clustering[tr1+'_'+tr2] = calc_wp_fast(x1, y1, z1, rpbins, pimax, self.lbox, Nthread, 
+                        x2 = x2, y2 = y2, z2 = z2)
+                    clustering[tr2+'_'+tr1] = clustering[tr1+'_'+tr2]
         return clustering
 
-    def gal_reader(self, scratch_dir = None, simname = None, 
+    def gal_reader(self, output_dir = None, simname = None, 
         sim_dir = None, z_mock = None, want_rsd = None, tracers = None):
         """
         Loads galaxy data given directory and return a ``mock_dict`` dictionary.  
@@ -692,7 +869,7 @@ class AbacusHOD:
         ``sim_dir``: str
             the directory that the simulation lives in, e.g. '/path/to/AbacusSummit/'.                                 
 
-        ``scratch_dir``: str
+        ``output_dir``: str
             the diretory to save galaxy to, e.g. '/my/output/galalxies'. 
 
         ``z_mock``: floa
@@ -711,8 +888,8 @@ class AbacusHOD:
 
         """
 
-        if scratch_dir == None:
-            scratch_dir = Path(self.scratch_dir)
+        if output_dir == None:
+            output_dir = Path(self.output_dir)
         if simname == None:
             simname = Path(self.sim_name)
         if sim_dir == None:
@@ -723,7 +900,7 @@ class AbacusHOD:
             want_rsd = self.want_rsd
         if tracers == None:
             tracers = self.tracers.keys()
-        mock_dir = scratch_dir / simname / ('z%4.3f'%self.z_mock)
+        mock_dir = output_dir / simname / ('z%4.3f'%self.z_mock)
 
         if want_rsd:
             rsd_string = "_rsd"
